from __future__ import annotations

import json
from typing import TYPE_CHECKING

from mars_patcher.compress import comp_lz77, decomp_lz77
from mars_patcher.constants.game_data import minimap_ptrs
from mars_patcher.data import get_data_path

if TYPE_CHECKING:
    from types import TracebackType

    from mars_patcher.common_types import MinimapId
    from mars_patcher.rom import Rom

MINIMAP_DIM = 32


class Minimap:
    """Class for reading/writing minimap data and setting tiles."""

    def __init__(self, rom: Rom, id: MinimapId):
        self.rom = rom
        self.pointer = minimap_ptrs(rom) + (id * 4)
        addr = rom.read_ptr(self.pointer)
        self.tile_data, self.comp_len = decomp_lz77(rom.data, addr)

    def __enter__(self) -> Minimap:
        # We don't need to do anything
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.write()

    def get_tile_value(self, x: int, y: int) -> tuple[int, int, bool, bool]:
        idx = (y * MINIMAP_DIM + x) * 2
        if idx >= len(self.tile_data):
            raise IndexError(f"Tile coordinate ({x}, {y}) is not within minimap")
        value = self.tile_data[idx] | self.tile_data[idx + 1] << 8
        tile = value & (0x3FF)
        palette = value >> 12
        h_flip = value & 0x400 != 0
        v_flip = value & 0x800 != 0
        return tile, palette, h_flip, v_flip

    def set_tile_value(
        self, x: int, y: int, tile: int, palette: int, h_flip: bool = False, v_flip: bool = False
    ) -> None:
        idx = (y * MINIMAP_DIM + x) * 2
        if idx >= len(self.tile_data):
            raise IndexError(f"Tile coordinate ({x}, {y}) is not within minimap")
        value = tile | (palette << 12)
        if h_flip:
            value |= 0x400
        if v_flip:
            value |= 0x800
        self.tile_data[idx] = value & 0xFF
        self.tile_data[idx + 1] = value >> 8

    def write(self) -> None:
        comp_data = comp_lz77(self.tile_data)
        comp_len = len(comp_data)
        if comp_len > self.comp_len:
            # Repoint data
            addr = self.rom.reserve_free_space(comp_len)
            self.rom.write_ptr(self.pointer, addr)
        else:
            addr = self.rom.read_ptr(self.pointer)
        self.rom.write_bytes(addr, comp_data)
        self.comp_len = comp_len


def apply_minimap_edits(rom: Rom, edit_dict: dict) -> None:
    # Go through every minimap
    for map_id, changes in edit_dict.items():
        with Minimap(rom, int(map_id)) as minimap:
            for change in changes:
                minimap.set_tile_value(
                    change["X"],
                    change["Y"],
                    change["Tile"],
                    change["Palette"],
                    change.get("HFlip", False),
                    change.get("VFlip", False),
                )


def apply_base_minimap_edits(rom: Rom) -> None:
    with open(get_data_path("base_minimap_edits.json")) as f:
        data = json.load(f)

    # Go through every minimap
    for map in data:
        with Minimap(rom, int(map["MAP_ID"])) as minimap:
            for change in map["CHANGES"]:
                minimap.set_tile_value(
                    change["X"],
                    change["Y"],
                    change["Tile"],
                    change["Palette"],
                    change.get("HFlip", False),
                    change.get("VFlip", False),
                )



from __future__ import annotations

from enum import Enum
from typing import NamedTuple

from typing_extensions import Self

# String Format
# <top><left><right><bottom>_<top-left><top-right><bottom-left><bottom-right>_<content>

# Chunk 1 (tile edges)
# - see Edge and ColoredDoor for values

# Chunk 2 (tile corners)
# - C: corner pixel
# - x: none

# Chunk 3 (tile content)
# - see Content for values


# Edges
class Edge(Enum):
    EMPTY = "x"
    WALL = "W"
    SHORTCUT = "S"
    DOOR = "D"

    @property
    def is_door(self) -> bool:
        return self == Edge.DOOR


class ColoredDoor(Enum):
    BLUE = "B"
    GREEN = "G"
    YELLOW = "Y"
    RED = "R"

    @property
    def is_door(self) -> bool:
        return True

    # Aliases
    L1 = BLUE
    L2 = GREEN
    L3 = YELLOW
    L4 = RED


class TileEdges(NamedTuple):
    top: Edge = Edge.WALL
    left: Edge | ColoredDoor = Edge.WALL
    right: Edge | ColoredDoor = Edge.WALL
    bottom: Edge = Edge.WALL

    @property
    def as_str(self) -> str:
        return f"{self.top.value}{self.left.value}{self.right.value}{self.bottom.value}"

    @classmethod
    def from_str(cls, value: str) -> Self:
        if len(value) != 4:
            raise ValueError(f"'{value}' is not a valid TileEdges string")
        top, left, right, bottom = tuple(value)

        def any_edge_from_value(v: str) -> Edge | ColoredDoor:
            try:
                return Edge(v)
            except ValueError:
                pass

            try:
                return ColoredDoor(v)
            except ValueError:
                raise ValueError(f"{repr(v)} is not a valid Edge or ColoredDoor")

        return cls(
            top=Edge(top),
            left=any_edge_from_value(left),
            right=any_edge_from_value(right),
            bottom=Edge(bottom),
        )

    def h_flip(self) -> TileEdges:
        return TileEdges(
            top=self.top,
            left=self.right,
            right=self.left,
            bottom=self.bottom,
        )

    def v_flip(self) -> TileEdges:
        return TileEdges(
            top=self.bottom,
            left=self.left,
            right=self.right,
            bottom=self.top,
        )


# Corners
class TileCorners(NamedTuple):
    top_left: bool = False
    top_right: bool = False
    bottom_left: bool = False
    bottom_right: bool = False

    @property
    def as_str(self) -> str:
        def s(corner: bool) -> str:
            return "C" if corner else "x"

        return f"{s(self.top_left)}{s(self.top_right)}{s(self.bottom_left)}{s(self.bottom_right)}"

    @classmethod
    def from_str(cls, value: str) -> Self:
        if len(value) != 4:
            raise ValueError(f"'{value}' is not a valid TileCorners string")
        tl, tr, bl, br = tuple(value)
        return cls(
            top_left=(tl == "C"),
            top_right=(tr == "C"),
            bottom_left=(bl == "C"),
            bottom_right=(br == "C"),
        )

    def h_flip(self) -> TileCorners:
        return TileCorners(
            top_left=self.top_right,
            top_right=self.top_left,
            bottom_left=self.bottom_right,
            bottom_right=self.bottom_left,
        )

    def v_flip(self) -> TileCorners:
        return TileCorners(
            top_left=self.bottom_left,
            top_right=self.bottom_right,
            bottom_left=self.top_left,
            bottom_right=self.top_right,
        )


# Contents
class Content(Enum):
    EMPTY = "x"
    NAVIGATION = "N"
    SAVE = "S"
    RECHARGE = "R"
    HIDDEN_RECHARGE = "H"
    DATA = "D"
    ITEM = "I"
    OBTAINED_ITEM = "O"
    BOSS = "B"
    GUNSHIP = "G"
    GUNSHIP_EDGE = "P"
    SECURITY = "K"
    AUXILLARY_POWER = "X"
    ANIMALS = "A"
    BOILER_PAD = "b"
    TUNNEL = "T"

    @property
    def can_h_flip(self) -> bool:
        exclude = {
            Content.NAVIGATION,
            Content.SAVE,
            Content.RECHARGE,
            Content.HIDDEN_RECHARGE,
            Content.DATA,
            Content.GUNSHIP,
            Content.GUNSHIP_EDGE,
        }
        return self not in exclude

    @property
    def can_v_flip(self) -> bool:
        exclude = {
            Content.NAVIGATION,
            Content.SAVE,
            Content.RECHARGE,
            Content.HIDDEN_RECHARGE,
            Content.GUNSHIP,
            Content.GUNSHIP_EDGE,
            Content.BOSS,
        }
        return self not in exclude


# Tile
class MapTile(NamedTuple):
    edges: TileEdges = TileEdges()
    corners: TileCorners = TileCorners()
    content: Content = Content.EMPTY

    @property
    def as_str(self) -> str:
        return f"{self.edges.as_str}_{self.corners.as_str}_{self.content.value}"

    @classmethod
    def from_str(cls, value: str) -> Self:
        if len(value) != 11:
            raise ValueError(f"'{value}' is not a valid MapTile string")
        edges, corners, content = value.split("_")
        return cls(
            edges=TileEdges.from_str(edges),
            corners=TileCorners.from_str(corners),
            content=Content(content),
        )

    def h_flip(self) -> MapTile:
        if not self.content.can_h_flip:
            raise ValueError(f"Cannot h_flip tile with contents {self.content}")

        return MapTile(
            edges=self.edges.h_flip(),
            corners=self.corners.h_flip(),
            content=self.content,
        )

    def v_flip(self) -> MapTile:
        if not self.content.can_v_flip:
            raise ValueError(f"Cannot v_flip tile with contents {self.content}")

        return MapTile(
            edges=self.edges.v_flip(),
            corners=self.corners.v_flip(),
            content=self.content,
        )


# Constants
COLORED_DOOR_TILES = {
    0x005: MapTile.from_str("WBxx_xxxx_x"),
    0x006: MapTile.from_str("DBxx_xxxx_x"),
    0x007: MapTile.from_str("DBxx_xxxC_x"),
    0x008: MapTile.from_str("DBBD_xxxx_x"),
    0x009: MapTile.from_str("DBBx_xxxx_x"),
    0x025: MapTile.from_str("xBxx_xxxx_x"),
    0x026: MapTile.from_str("xBxx_xCxC_x"),
    0x027: MapTile.from_str("DBWD_xxxx_x"),
    0x028: MapTile.from_str("DBDx_xxxx_x"),
    0x029: MapTile.from_str("DBBW_xxxx_x"),
    0x045: MapTile.from_str("WBxW_xxxx_x"),
    0x046: MapTile.from_str("DBxD_xxxx_x"),
    0x047: MapTile.from_str("DBxW_xxxx_x"),
    0x048: MapTile.from_str("WBWW_xxxx_x"),
    0x049: MapTile.from_str("WBBW_xxxx_x"),
    0x065: MapTile.from_str("DBWx_xxxx_x"),
    0x066: MapTile.from_str("WBDx_xxxx_x"),
    0x068: MapTile.from_str("WBWx_xxxx_x"),
    0x069: MapTile.from_str("WBBx_xxxx_x"),
    0x085: MapTile.from_str("DBWW_xxxx_x"),
    0x086: MapTile.from_str("xBDx_xxxx_x"),
    0x088: MapTile.from_str("xBWx_xxxx_x"),
    0x089: MapTile.from_str("xBBx_xxxx_x"),
    0x00A: MapTile.from_str("WGxx_xxxx_x"),
    0x00B: MapTile.from_str("DGxx_xxxx_x"),
    0x00C: MapTile.from_str("DGxx_xxxC_x"),
    0x00D: MapTile.from_str("DGGD_xxxx_x"),
    0x00E: MapTile.from_str("DGGx_xxxx_x"),
    0x02A: MapTile.from_str("xGxx_xxxx_x"),
    0x02B: MapTile.from_str("xGxx_xCxC_x"),
    0x02C: MapTile.from_str("DGWD_xxxx_x"),
    0x02D: MapTile.from_str("DGDx_xxxx_x"),
    0x02E: MapTile.from_str("DGGW_xxxx_x"),
    0x04A: MapTile.from_str("WGxW_xxxx_x"),
    0x04B: MapTile.from_str("DGxD_xxxx_x"),
    0x04C: MapTile.from_str("DGxW_xxxx_x"),
    0x04D: MapTile.from_str("WGWW_xxxx_x"),
    0x04E: MapTile.from_str("WGGW_xxxx_x"),
    0x06A: MapTile.from_str("DGWx_xxxx_x"),
    0x06B: MapTile.from_str("WGDx_xxxx_x"),
    0x06D: MapTile.from_str("WGWx_xxxx_x"),
    0x06E: MapTile.from_str("WGGx_xxxx_x"),
    0x08A: MapTile.from_str("DGWW_xxxx_x"),
    0x08B: MapTile.from_str("xGDx_xxxx_x"),
    0x08D: MapTile.from_str("xGWx_xxxx_x"),
    0x08E: MapTile.from_str("xGGx_xxxx_x"),
    0x00F: MapTile.from_str("WRxx_xxxx_x"),
    0x010: MapTile.from_str("DRxx_xxxx_x"),
    0x011: MapTile.from_str("DRxx_xxxC_x"),
    0x012: MapTile.from_str("DRRD_xxxx_x"),
    0x013: MapTile.from_str("DRRx_xxxx_x"),
    0x02F: MapTile.from_str("xRxx_xxxx_x"),
    0x030: MapTile.from_str("xRxx_xCxC_x"),
    0x031: MapTile.from_str("DRWD_xxxx_x"),
    0x032: MapTile.from_str("DRDx_xxxx_x"),
    0x033: MapTile.from_str("DRRW_xxxx_x"),
    0x04F: MapTile.from_str("WRxW_xxxx_x"),
    0x050: MapTile.from_str("DRxD_xxxx_x"),
    0x051: MapTile.from_str("DRxW_xxxx_x"),
    0x052: MapTile.from_str("WRWW_xxxx_x"),
    0x053: MapTile.from_str("WRRW_xxxx_x"),
    0x06F: MapTile.from_str("DRWx_xxxx_x"),
    0x070: MapTile.from_str("WRDx_xxxx_x"),
    0x072: MapTile.from_str("WRWx_xxxx_x"),
    0x073: MapTile.from_str("WRRx_xxxx_x"),
    0x08F: MapTile.from_str("DRWW_xxxx_x"),
    0x090: MapTile.from_str("xRDx_xxxx_x"),
    0x092: MapTile.from_str("xRWx_xxxx_x"),
    0x093: MapTile.from_str("xRRx_xxxx_x"),
    0x014: MapTile.from_str("WYxx_xxxx_x"),
    0x015: MapTile.from_str("DYxx_xxxx_x"),
    0x016: MapTile.from_str("DYxx_xxxC_x"),
    0x017: MapTile.from_str("DYYD_xxxx_x"),
    0x018: MapTile.from_str("DYYx_xxxx_x"),
    0x034: MapTile.from_str("xYxx_xxxx_x"),
    0x035: MapTile.from_str("xYxx_xCxC_x"),
    0x036: MapTile.from_str("DYWD_xxxx_x"),
    0x037: MapTile.from_str("DYDx_xxxx_x"),
    0x038: MapTile.from_str("DYYW_xxxx_x"),
    0x054: MapTile.from_str("WYxW_xxxx_x"),
    0x055: MapTile.from_str("DYxD_xxxx_x"),
    0x056: MapTile.from_str("DYxW_xxxx_x"),
    0x057: MapTile.from_str("WYWW_xxxx_x"),
    0x058: MapTile.from_str("WYYW_xxxx_x"),
    0x074: MapTile.from_str("DYWx_xxxx_x"),
    0x075: MapTile.from_str("WYDx_xxxx_x"),
    0x077: MapTile.from_str("WYWx_xxxx_x"),
    0x078: MapTile.from_str("WYYx_xxxx_x"),
    0x094: MapTile.from_str("DYWW_xxxx_x"),
    0x095: MapTile.from_str("xYDx_xxxx_x"),
    0x097: MapTile.from_str("xYWx_xxxx_x"),
    0x098: MapTile.from_str("xYYx_xxxx_x"),
    0x019: MapTile.from_str("DBGx_xxxx_x"),
    0x01A: MapTile.from_str("DBRx_xxxx_x"),
    0x01B: MapTile.from_str("DBYx_xxxx_x"),
    0x01C: MapTile.from_str("DGRx_xxxx_x"),
    0x01D: MapTile.from_str("DGYx_xxxx_x"),
    0x01E: MapTile.from_str("DRYx_xxxx_x"),
    0x039: MapTile.from_str("DBGW_xxxx_x"),
    0x03A: MapTile.from_str("DBRW_xxxx_x"),
    0x03B: MapTile.from_str("DBYW_xxxx_x"),
    0x03C: MapTile.from_str("DGRW_xxxx_x"),
    0x03D: MapTile.from_str("DGYW_xxxx_x"),
    0x03E: MapTile.from_str("DRYW_xxxx_x"),
    0x059: MapTile.from_str("WBGW_xxxx_x"),
    0x05A: MapTile.from_str("WBRW_xxxx_x"),
    0x05B: MapTile.from_str("WBYW_xxxx_x"),
    0x05C: MapTile.from_str("WGRW_xxxx_x"),
    0x05D: MapTile.from_str("WGYW_xxxx_x"),
    0x05E: MapTile.from_str("WRYW_xxxx_x"),
    0x079: MapTile.from_str("WBGx_xxxx_x"),
    0x07A: MapTile.from_str("WBRx_xxxx_x"),
    0x07B: MapTile.from_str("WBYx_xxxx_x"),
    0x07C: MapTile.from_str("WGRx_xxxx_x"),
    0x07D: MapTile.from_str("WGYx_xxxx_x"),
    0x07E: MapTile.from_str("WRYx_xxxx_x"),
    0x099: MapTile.from_str("xBGx_xxxx_x"),
    0x09A: MapTile.from_str("xBRx_xxxx_x"),
    0x09B: MapTile.from_str("xBYx_xxxx_x"),
    0x09C: MapTile.from_str("xGRx_xxxx_x"),
    0x09D: MapTile.from_str("xGYx_xxxx_x"),
    0x09E: MapTile.from_str("xRYx_xxxx_x"),
    0x01F: MapTile.from_str("DBGD_xxxx_x"),
    0x03F: MapTile.from_str("DBRD_xxxx_x"),
    0x05F: MapTile.from_str("DBYD_xxxx_x"),
    0x07F: MapTile.from_str("DGRD_xxxx_x"),
    0x09F: MapTile.from_str("DGYD_xxxx_x"),
    0x096: MapTile.from_str("DRYD_xxxx_x"),
    0x0E6: MapTile.from_str("WDBW_xxxx_x"),
    0x0E7: MapTile.from_str("DDBW_xxxx_x"),
    0x0E8: MapTile.from_str("DDBD_xxxx_x"),
    0x0EA: MapTile.from_str("DGxx_xxxC_x"),
    0x0EB: MapTile.from_str("WDGW_xxxx_x"),
    0x0EC: MapTile.from_str("DDGW_xxxx_x"),
    0x0ED: MapTile.from_str("DDGD_xxxx_x"),
    0x0EF: MapTile.from_str("DRxx_xxxC_x"),
    0x0F0: MapTile.from_str("WDRW_xxxx_x"),
    0x0F1: MapTile.from_str("DDRW_xxxx_x"),
    0x0F2: MapTile.from_str("DDRD_xxxx_x"),
    0x0F4: MapTile.from_str("DYxx_xxxC_x"),
    0x0F5: MapTile.from_str("WDYW_xxxx_x"),
    0x0F6: MapTile.from_str("DDYW_xxxx_x"),
    0x0F7: MapTile.from_str("DDYD_xxxx_x"),
    0x0F8: MapTile.from_str("WYxx_xxxC_x"),
    0x06C: MapTile.from_str("WGxx_xxxC_x"),
    0x140: MapTile.from_str("WBWW_xxxx_R"),
    0x141: MapTile.from_str("WWBW_xxxx_R"),
    0x142: MapTile.from_str("WBBW_xxxx_R"),
    0x14C: MapTile.from_str("WBWW_xxxx_N"),
    0x14D: MapTile.from_str("WWBW_xxxx_N"),
    0x14E: MapTile.from_str("WBBW_xxxx_N"),
    0x160: MapTile.from_str("WBWW_xxxx_D"),
    0x161: MapTile.from_str("WWBW_xxxx_D"),
    0x162: MapTile.from_str("WBBW_xxxx_D"),
    0x143: MapTile.from_str("WGWW_xxxx_R"),
    0x144: MapTile.from_str("WWGW_xxxx_R"),
    0x145: MapTile.from_str("WGGW_xxxx_R"),
    0x14F: MapTile.from_str("WGWW_xxxx_N"),
    0x150: MapTile.from_str("WWGW_xxxx_N"),
    0x151: MapTile.from_str("WGGW_xxxx_N"),
    0x163: MapTile.from_str("WGWW_xxxx_D"),
    0x164: MapTile.from_str("WWGW_xxxx_D"),
    0x165: MapTile.from_str("WGGW_xxxx_D"),
    0x146: MapTile.from_str("WRWW_xxxx_R"),
    0x147: MapTile.from_str("WWRW_xxxx_R"),
    0x148: MapTile.from_str("WRRW_xxxx_R"),
    0x152: MapTile.from_str("WRWW_xxxx_N"),
    0x153: MapTile.from_str("WWRW_xxxx_N"),
    0x154: MapTile.from_str("WRRW_xxxx_N"),
    0x166: MapTile.from_str("WRWW_xxxx_D"),
    0x167: MapTile.from_str("WWRW_xxxx_D"),
    0x168: MapTile.from_str("WRRW_xxxx_D"),
    0x149: MapTile.from_str("WYWW_xxxx_R"),
    0x14A: MapTile.from_str("WWYW_xxxx_R"),
    0x14B: MapTile.from_str("WYYW_xxxx_R"),
    0x155: MapTile.from_str("WYWW_xxxx_N"),
    0x156: MapTile.from_str("WWYW_xxxx_N"),
    0x157: MapTile.from_str("WYYW_xxxx_N"),
    0x169: MapTile.from_str("WYWW_xxxx_D"),
    0x16A: MapTile.from_str("WWYW_xxxx_D"),
    0x16B: MapTile.from_str("WYYW_xxxx_D"),
    0x128: MapTile.from_str("WRDW_xxxx_R"),
    0x15E: MapTile.from_str("WDYW_xxxx_R"),
    0x17E: MapTile.from_str("WRRW_xxxx_S"),
    0x17F: MapTile.from_str("WYDW_xxxx_S"),
    0x198: MapTile.from_str("WBWW_xxxx_I"),
    0x199: MapTile.from_str("WBWW_xxxx_O"),
    0x19E: MapTile.from_str("WGWW_xxxx_I"),
    0x19F: MapTile.from_str("WGWW_xxxx_O"),
    0x1AC: MapTile.from_str("WYYW_xxxx_I"),
    0x1AD: MapTile.from_str("WYYW_xxxx_O"),
    # New Tiles
    0x10B: MapTile.from_str("xxGW_Cxxx_B"),
    0x16C: MapTile.from_str("xWBW_xxxx_K"),
    0x16D: MapTile.from_str("WWGW_xxxx_K"),
    0x16E: MapTile.from_str("WWRW_xxxx_K"),
    0x16F: MapTile.from_str("WYDW_xxxx_K"),
    0x1AE: MapTile.from_str("WYWW_xxxx_I"),
    0x1AF: MapTile.from_str("WYWW_xxxx_O"),
}


NORMAL_DOOR_TILES = {
    0x000: MapTile.from_str("WWxx_xxxx_x"),
    0x001: MapTile.from_str("Wxxx_xxxx_x"),
    0x002: MapTile.from_str("DWxx_xxxx_x"),
    0x003: MapTile.from_str("Dxxx_xxxx_x"),
    0x004: MapTile.from_str("DWxx_xxxC_x"),
    0x020: MapTile.from_str("xWxx_xxxx_x"),
    0x021: MapTile.from_str("xDxx_xxxx_x"),
    0x022: MapTile.from_str("Wxxx_xxCC_x"),
    0x023: MapTile.from_str("xWxx_xCxC_x"),
    0x024: MapTile.from_str("DWWD_xxxx_x"),
    0x040: MapTile.from_str("WWxW_xxxx_x"),
    0x041: MapTile.from_str("WxxW_xxxx_x"),
    0x042: MapTile.from_str("DWxD_xxxx_x"),
    0x043: MapTile.from_str("DWxW_xxxx_x"),
    0x044: MapTile.from_str("DxxD_xxxx_x"),
    0x060: MapTile.from_str("DWWx_xxxx_x"),
    0x061: MapTile.from_str("WDWx_xxxx_x"),
    0x062: MapTile.from_str("WDDx_xxxx_x"),
    0x063: MapTile.from_str("WWWx_xxxx_x"),
    0x064: MapTile.from_str("DxxW_xxxx_x"),
    0x080: MapTile.from_str("DWWW_xxxx_x"),
    0x081: MapTile.from_str("xDWx_xxxx_x"),
    0x082: MapTile.from_str("xDDx_xxxx_x"),
    0x083: MapTile.from_str("xWWx_xxxx_x"),
    0x084: MapTile.from_str("WDDW_xxxx_x"),
    0x087: MapTile.from_str("WWWW_xxxx_x"),
    0x0B4: MapTile.from_str("WSDW_xxxx_I"),
    0x0B6: MapTile.from_str("WSDW_xxxx_x"),
    0x0E0: MapTile.from_str("DDxW_xxxx_x"),
    0x0E1: MapTile.from_str("DDxD_xxxx_x"),
    0x0E2: MapTile.from_str("DDWx_xxxx_x"),
    0x0E3: MapTile.from_str("DDDx_xxxx_x"),
    0x0E4: MapTile.from_str("WWxx_xxxC_x"),
    0x0E5: MapTile.from_str("xWxx_xxxC_x"),
    0x100: MapTile.from_str("Dxxx_xxCC_x"),
    0x101: MapTile.from_str("xDxx_xCxC_x"),
    0x102: MapTile.from_str("DDxx_xxxC_x"),
    0x104: MapTile.from_str("WDxx_xxxC_x"),
    0x105: MapTile.from_str("xDxx_xCxx_x"),
    0x108: MapTile.from_str("xxDW_xxxx_B"),
    0x109: MapTile.from_str("WDxW_xxxx_B"),
    0x10D: MapTile.from_str("xxDx_xxxx_B"),
    0x10F: MapTile.from_str("xDxW_xxxx_B"),
    0x120: MapTile.from_str("DDWW_xxxx_x"),
    0x121: MapTile.from_str("DDDW_xxxx_x"),
    0x122: MapTile.from_str("DDWD_xxxx_x"),
    0x123: MapTile.from_str("DDDD_xxxx_x"),
    0x124: MapTile.from_str("WDWW_xxxx_x"),
    0x125: MapTile.from_str("WDxx_xxxx_x"),
    0x126: MapTile.from_str("xxxx_xxxx_x"),
    0x127: MapTile.from_str("Wxxx_xxxC_x"),
    0x129: MapTile.from_str("WDxW_xxxx_P"),
    0x12A: MapTile.from_str("WxWW_xxxx_G"),
    0x12B: MapTile.from_str("WDxW_xxxx_x"),
    0x12C: MapTile.from_str("DDxx_xxxx_x"),
    0x138: MapTile.from_str("WDWW_xxxx_H"),
    0x139: MapTile.from_str("WWDW_xxxx_H"),
    0x13A: MapTile.from_str("WDDW_xxxx_H"),
    0x158: MapTile.from_str("WDWW_xxxx_R"),
    0x159: MapTile.from_str("WWDW_xxxx_R"),
    0x15A: MapTile.from_str("WDDW_xxxx_R"),
    0x15B: MapTile.from_str("WDWW_xxxx_D"),
    0x15C: MapTile.from_str("WWDW_xxxx_D"),
    0x15D: MapTile.from_str("WDDW_xxxx_D"),
    0x178: MapTile.from_str("WDWW_xxxx_N"),
    0x179: MapTile.from_str("WWDW_xxxx_N"),
    0x17A: MapTile.from_str("WDDW_xxxx_N"),
    0x17B: MapTile.from_str("WDWW_xxxx_S"),
    0x17C: MapTile.from_str("WWDW_xxxx_S"),
    0x17D: MapTile.from_str("WDDW_xxxx_S"),
    0x180: MapTile.from_str("WDWW_xxxx_I"),
    0x181: MapTile.from_str("WDWW_xxxx_O"),
    0x182: MapTile.from_str("WWxW_xxxx_I"),
    0x183: MapTile.from_str("WWxW_xxxx_O"),
    0x184: MapTile.from_str("WDDW_xxxx_I"),
    0x185: MapTile.from_str("WDDW_xxxx_O"),
    0x186: MapTile.from_str("WWxx_xxxx_I"),
    0x187: MapTile.from_str("WWxx_xxxx_O"),
    0x188: MapTile.from_str("WDxW_xxxx_I"),
    0x189: MapTile.from_str("WDxW_xxxx_O"),
    0x18A: MapTile.from_str("xWxx_xxxx_I"),
    0x18B: MapTile.from_str("xWxx_xxxx_O"),
    0x18C: MapTile.from_str("xDxx_xxxx_I"),
    0x18D: MapTile.from_str("xDxx_xxxx_O"),
    0x18E: MapTile.from_str("WDxx_xxxx_I"),
    0x18F: MapTile.from_str("WDxx_xxxx_O"),
    0x190: MapTile.from_str("Wxxx_xxxx_I"),
    0x191: MapTile.from_str("Wxxx_xxxx_O"),
    0x192: MapTile.from_str("WxxW_xxxx_I"),
    0x193: MapTile.from_str("WxxW_xxxx_O"),
    0x194: MapTile.from_str("WWWx_xxxx_I"),
    0x195: MapTile.from_str("WWWx_xxxx_O"),
    0x196: MapTile.from_str("DDDW_xxxx_I"),
    0x197: MapTile.from_str("DDDW_xxxx_O"),
    0x19A: MapTile.from_str("xDDW_xxxx_I"),
    0x19B: MapTile.from_str("xDDW_xxxx_O"),
    0x19C: MapTile.from_str("WWDx_xxxx_I"),
    0x19D: MapTile.from_str("WWDx_xxxx_O"),
    0x1A0: MapTile.from_str("xWWD_xxxx_I"),
    0x1A1: MapTile.from_str("xWWD_xxxx_O"),
    0x1A2: MapTile.from_str("WWWD_xxxx_I"),
    0x1A3: MapTile.from_str("WWWD_xxxx_O"),
    0x1A4: MapTile.from_str("xxxx_xxxx_I"),
    0x1A5: MapTile.from_str("xxxx_xxxx_O"),
    0x1A6: MapTile.from_str("WWxD_xxxx_I"),
    0x1A7: MapTile.from_str("WWxD_xxxx_O"),
    0x1A8: MapTile.from_str("xWWx_xxxx_I"),
    0x1A9: MapTile.from_str("xWWx_xxxx_O"),
    0x1AA: MapTile.from_str("DDWW_xxxx_I"),
    0x1AB: MapTile.from_str("DDWW_xxxx_O"),
    0x0B5: MapTile.from_str("WSDW_xxxx_O"),
    # New Tiles
    0x0A4: MapTile.from_str("WDxW_xxxx_T"),
    0x0D4: MapTile.from_str("WSDW_xxxx_I"),
    0x0D5: MapTile.from_str("WSDW_xxxx_O"),
    0x10A: MapTile.from_str("WxDx_xxxx_B"),
    0x10C: MapTile.from_str("xDxW_xCxx_B"),
    0x170: MapTile.from_str("WWDx_xxxx_x"),
    0x172: MapTile.from_str("WDDW_xxxx_X"),
}

COLORED_DOOR_TILE_IDS = {tile: id for id, tile in COLORED_DOOR_TILES.items()}
NORMAL_DOOR_TILE_IDS = {tile: id for id, tile in NORMAL_DOOR_TILES.items()}

ALL_DOOR_TILES = COLORED_DOOR_TILES | NORMAL_DOOR_TILES
ALL_DOOR_TILE_IDS = COLORED_DOOR_TILE_IDS | NORMAL_DOOR_TILE_IDS



import math
from enum import Enum
from typing import Any


class RgbBitSize(Enum):
    Rgb5 = 1
    Rgb8 = 2


class RgbColor:
    """Color represented as RGB using 5 or 8 bits per channel."""

    FACTOR = 255.0

    def __init__(self, R: int, G: int, B: int, bit_size: RgbBitSize):
        if bit_size == RgbBitSize.Rgb5:
            R <<= 3
            G <<= 3
            B <<= 3
        elif bit_size == RgbBitSize.Rgb8:
            self.red = R
            self.green = G
            self.blue = B
        else:
            raise ValueError(bit_size)
        self.red = R
        self.green = G
        self.blue = B

    @classmethod
    def from_rgb(cls, rgb: int, bit_size: RgbBitSize) -> "RgbColor":
        if bit_size == RgbBitSize.Rgb5:
            r = (rgb & 0x1F) << 3
            g = (rgb & 0x3E0) >> 2
            b = (rgb & 0x7C00) >> 7
        elif bit_size == RgbBitSize.Rgb8:
            r = (rgb >> 16) & 0xFF
            g = (rgb >> 8) & 0xFF
            b = rgb & 0xFF
        else:
            raise ValueError(bit_size)
        return RgbColor(r, g, b, RgbBitSize.Rgb8)

    def __str__(self) -> str:
        return self.hex_15()

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, RgbColor):
            return self.rgb_24() == other.rgb_24()
        return False

    def __hash__(self) -> int:
        return self.rgb_24()

    def hsv(self) -> "HsvColor":
        r = self.r_fraction()
        g = self.g_fraction()
        b = self.b_fraction()
        channels = [r, g, b]
        channels.sort()

        # Get value
        c_min = channels[0]
        c_max = channels[2]
        v = c_max

        # Get hue
        c_range = c_max - c_min
        if c_range == 0:
            h = 0.0
        else:
            if c_max == r:
                h = 60 * ((g - b) / c_range)
            elif c_max == g:
                h = 60 * ((b - r) / c_range + 2)
            elif c_max == b:
                h = 60 * ((r - g) / c_range + 4)
            # Check if negative
            if h < 0:
                h += 360

        # Get saturation
        if v == 0:
            s = 0.0
        else:
            s = c_range / v

        return HsvColor(h, s, v)

    def oklab(self) -> "OklabColor":
        # Convert to linear RGB
        rl = self.srgb_to_linear(self.r_fraction())
        gl = self.srgb_to_linear(self.g_fraction())
        bl = self.srgb_to_linear(self.b_fraction())

        # Convert to LMS
        lg = 0.4122214708 * rl + 0.5363325363 * gl + 0.0514459929 * bl
        md = 0.2119034982 * rl + 0.6806995451 * gl + 0.1073969566 * bl
        st = 0.0883024619 * rl + 0.2817188376 * gl + 0.6299787005 * bl

        lg = lg ** (1 / 3)
        md = md ** (1 / 3)
        st = st ** (1 / 3)

        return OklabColor(
            0.2104542553 * lg + 0.7936177850 * md - 0.0040720468 * st,
            1.9779984951 * lg - 2.4285922050 * md + 0.4505937099 * st,
            0.0259040371 * lg + 0.7827717662 * md - 0.8086757660 * st,
        )

    def luma(self) -> float:
        return 0.299 * self.red + 0.587 * self.green + 0.114 * self.blue

    def r_5(self) -> int:
        return self.red >> 3

    def g_5(self) -> int:
        return self.green >> 3

    def b_5(self) -> int:
        return self.blue >> 3

    def rgb_15(self) -> int:
        return (self.b_5() << 10) | (self.g_5() << 5) | self.r_5()

    def rgb_24(self) -> int:
        return (self.blue << 16) | (self.green << 8) | self.red

    def hex_15(self) -> str:
        return f"{self.rgb_15():04X}"

    def r_fraction(self) -> float:
        return self.red / self.FACTOR

    def g_fraction(self) -> float:
        return self.green / self.FACTOR

    def b_fraction(self) -> float:
        return self.blue / self.FACTOR

    @classmethod
    def black(cls) -> "RgbColor":
        return RgbColor.from_rgb(0, RgbBitSize.Rgb8)

    @classmethod
    def white_5(cls) -> "RgbColor":
        return RgbColor.from_rgb(0x7FFF, RgbBitSize.Rgb5)

    @staticmethod
    def srgb_to_linear(value: float) -> float:
        if value > 0.04045:
            return math.pow((value + 0.055) / 1.055, 2.4)
        return value / 12.92


class HsvColor:
    """
    Color represented as HSV, where 0 <= hue <= 360,
    0 <= saturation <= 1, and 0 <= value <= 1.
    See https://en.wikipedia.org/wiki/HSL_and_HSV
    """

    def __init__(self, hue: float, saturation: float, value: float):
        self.hue = hue
        self.saturation = saturation
        self.value = value

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, HsvColor):
            return (
                self.hue == other.hue
                and self.saturation == other.saturation
                and self.value == other.value
            )
        return False

    def __hash__(self) -> int:
        return hash(self.hue) ^ hash(self.saturation) ^ hash(self.value)

    def rgb(self) -> RgbColor:
        c = self.value * self.saturation
        hp = self.hue / 60
        x = c * (1 - abs(hp % 2 - 1))

        if hp < 1:
            rgb = (c, x, 0.0)
        elif hp < 2:
            rgb = (x, c, 0.0)
        elif hp < 3:
            rgb = (0.0, c, x)
        elif hp < 4:
            rgb = (0.0, x, c)
        elif hp < 5:
            rgb = (x, 0.0, c)
        else:
            rgb = (c, 0.0, x)

        m = self.value - c
        factor = RgbColor.FACTOR
        r = round((rgb[0] + m) * factor)
        g = round((rgb[1] + m) * factor)
        b = round((rgb[2] + m) * factor)
        return RgbColor(r, g, b, RgbBitSize.Rgb8)


class OklabColor:
    """
    Color represented as Oklab, where 0.0 <= L <= 1.0,
    and A and B are typically between -0.4 and 0.4.
    See https://bottosson.github.io/posts/oklab/
    """

    def __init__(self, L: float, A: float, B: float):
        self.l_star = L
        self.a_star = A
        self.b_star = B

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, OklabColor):
            return (
                self.l_star == other.l_star
                and self.a_star == other.a_star
                and self.b_star == other.b_star
            )
        return False

    def __hash__(self) -> int:
        return hash(self.l_star) ^ hash(self.a_star) ^ hash(self.b_star)

    def rgb(self) -> RgbColor:
        # Convert to LMS
        lg = self.l_star + 0.3963377774 * self.a_star + 0.2158037573 * self.b_star
        md = self.l_star - 0.1055613458 * self.a_star - 0.0638541728 * self.b_star
        st = self.l_star - 0.0894841775 * self.a_star - 1.2914855480 * self.b_star

        lg = lg**3
        md = md**3
        st = st**3

        # Convert to linear RGB
        rl = +4.0767416621 * lg - 3.3077115913 * md + 0.2309699292 * st
        gl = -1.2684380046 * lg + 2.6097574011 * md - 0.3413193965 * st
        bl = -0.0041960863 * lg - 0.7034186147 * md + 1.7076147010 * st

        # Convert to sRGB
        rf = self.linear_to_srgb(rl)
        gf = self.linear_to_srgb(gl)
        bf = self.linear_to_srgb(bl)

        r = int(round(rf * 255))
        g = int(round(gf * 255))
        b = int(round(bf * 255))

        return RgbColor(
            max(0, min(r, 255)), max(0, min(g, 255)), max(0, min(b, 255)), RgbBitSize.Rgb8
        )

    def hue(self) -> float:
        """Gets the hue measured in radians. Ranges from -pi to pi."""
        return math.atan2(self.b_star, self.a_star)

    def chroma(self) -> float:
        """
        The intensity or purity of a color, i.e. how far it is from a
        neutral gray of the same lightness."""
        return math.sqrt(self.a_star * self.a_star + self.b_star * self.b_star)

    def shift_hue(self, shift: float) -> None:
        """Shifts hue by the provided amount, measured in radians."""
        # Get hue in range 0 to 2pi
        hue = self.hue() + math.pi
        hue = (hue + shift) % (2 * math.pi)
        # Put hue back in range -pi to pi
        hue -= math.pi
        # Get new A and B values
        chroma = self.chroma()
        self.a_star = chroma * math.cos(hue)
        self.b_star = chroma * math.sin(hue)

    @staticmethod
    def linear_to_srgb(value: float) -> float:
        if value > 0.0031308:
            return 1.055 * math.pow(value, 1.0 / 2.4) - 0.055
        return value * 12.92



from enum import Enum
from zlib import crc32


class BpsDecodeError(Enum):
    INVALID_BPS = 0
    INVALID_SOURCE = 1
    ALREADY_PATCHED = 2


class BpsDecoder:
    def error(self, err: BpsDecodeError) -> None:
        if err == BpsDecodeError.INVALID_BPS:
            msg = "Invalid BPS file"
        elif err == BpsDecodeError.INVALID_SOURCE:
            msg = "Invalid source file"
        elif err == BpsDecodeError.ALREADY_PATCHED:
            msg = "File already patched"
        raise ValueError(msg)

    def apply_patch(self, patch: bytes, source: bytes, ignore_checksum: bool = False) -> bytearray:
        self.patch = patch
        self.source = source

        # Header
        self.patch_idx = 0
        marker = bytes(self.read_8() for _ in range(4))
        if marker != b"BPS1":
            self.error(BpsDecodeError.INVALID_BPS)
        source_size = self.decode_int()
        target_size = self.decode_int()
        metadata_size = self.decode_int()
        # Ignore metadata
        self.patch_idx += metadata_size

        # Check checksums
        patch_size = len(self.patch)
        if patch_size < self.patch_idx + 12:
            self.error(BpsDecodeError.INVALID_BPS)
        footer_start = patch_size - 12
        if not ignore_checksum:
            source_checksum_expected = self.read_32(footer_start)
            target_checksum_expected = self.read_32(footer_start + 4)
            patch_checksum_expected = self.read_32(footer_start + 8)

            patch_checksum_actual = crc32(self.patch[:-4])
            if patch_checksum_expected != patch_checksum_actual:
                self.error(BpsDecodeError.INVALID_BPS)

            source_checksum_actual = crc32(self.source)
            if source_size != len(source) or source_checksum_expected != source_checksum_actual:
                if (
                    len(source) == target_size
                    and source_checksum_actual == target_checksum_expected
                ):
                    self.error(BpsDecodeError.ALREADY_PATCHED)
                self.error(BpsDecodeError.INVALID_SOURCE)

        # Actions
        output_offset = 0
        source_offset = 0
        target_offset = 0
        target = bytearray()
        while self.patch_idx < footer_start:
            num = self.decode_int()
            length = (num >> 2) + 1
            if output_offset + length > target_size:
                self.error(BpsDecodeError.INVALID_BPS)
            action = num & 3
            if action == 0:
                # Source read
                for _ in range(length):
                    target.append(source[output_offset])
                    output_offset += 1
            elif action == 1:
                # Target read
                for _ in range(length):
                    target.append(self.read_8())
                    output_offset += 1
            elif action == 2:
                # Source copy
                offset = self.decode_int()
                source_offset += (-1 if offset & 1 else 1) * (offset >> 1)
                for _ in range(length):
                    target.append(source[source_offset])
                    output_offset += 1
                    source_offset += 1
            elif action == 3:
                # Target copy
                offset = self.decode_int()
                target_offset += (-1 if offset & 1 else 1) * (offset >> 1)
                for _ in range(length):
                    target.append(target[target_offset])
                    output_offset += 1
                    target_offset += 1
        if self.patch_idx > footer_start:
            self.error(BpsDecodeError.INVALID_BPS)
        if not ignore_checksum:
            target_checksum_actual = crc32(target)
            if target_checksum_expected != target_checksum_actual:
                self.error(BpsDecodeError.INVALID_BPS)
        return target

    def read_8(self) -> int:
        val = self.patch[self.patch_idx]
        self.patch_idx += 1
        return val

    def read_32(self, idx: int) -> int:
        return (
            self.patch[idx]
            | (self.patch[idx + 1] << 8)
            | (self.patch[idx + 2] << 16)
            | (self.patch[idx + 3] << 24)
        )

    def decode_int(self) -> int:
        num = 0
        shift = 1
        while True:
            x = self.read_8()
            num += (x & 0x7F) * shift
            if x & 0x80 != 0:
                return num
            shift <<= 7
            num += shift


class IpsDecodeError(Enum):
    INVALID_IPS = 0
    ABRUPT_IPS_END = 1
    PAST_TARGET_END = 2
    MISSING_EOF = 3


class IpsDecoder:
    def error(self, err: IpsDecodeError, extra: str | None = None) -> None:
        if err == IpsDecodeError.INVALID_IPS:
            msg = "Invalid IPS file"
        elif err == IpsDecodeError.ABRUPT_IPS_END:
            msg = "Abrupt end to IPS file"
        elif err == IpsDecodeError.PAST_TARGET_END:
            msg = "Trying to patch data past end of file"
        elif err == IpsDecodeError.MISSING_EOF:
            msg = "Improperly terminated IPS file"
        if extra is not None:
            msg += ", " + extra
        raise ValueError(msg)

    def apply_patch(self, patch: bytes, target: bytearray) -> None:
        # Check signature
        patch_len = len(patch)
        if patch_len < 8 or patch[:5] != b"PATCH":
            self.error(IpsDecodeError.INVALID_IPS)

        # Records
        idx = 5
        while idx + 2 < patch_len:
            # Check EOF
            if patch[idx : idx + 3] == b"EOF":
                return

            # Get address and size
            addr = (patch[idx] << 16) | (patch[idx + 1] << 8) | patch[idx + 2]
            idx += 3
            if idx + 1 >= patch_len:
                self.error(IpsDecodeError.ABRUPT_IPS_END, "entry cut off before size")
            size = (patch[idx] << 8) | patch[idx + 1]
            idx += 2

            if size == 0:
                # RLE
                if idx + 1 >= patch_len:
                    self.error(IpsDecodeError.ABRUPT_IPS_END, "entry cut off before RLE size")
                rle_size = (patch[idx] << 8) | patch[idx + 1]
                if addr + rle_size > len(target):
                    self.error(IpsDecodeError.PAST_TARGET_END)
                idx += 2
                if idx >= patch_len:
                    self.error(IpsDecodeError.ABRUPT_IPS_END, "entry cut off before RLE byte")
                rle_byte = patch[idx]
                idx += 1
                for i in range(addr, addr + rle_size):
                    target[i] = rle_byte
            else:
                if idx + size > patch_len:
                    self.error(
                        IpsDecodeError.ABRUPT_IPS_END, "entry cut off before end of data block"
                    )
                if addr + size > len(target):
                    self.error(IpsDecodeError.PAST_TARGET_END)
                target[addr : addr + size] = patch[idx : idx + size]
                idx += size

        self.error(IpsDecodeError.MISSING_EOF)



from mars_patcher.rom import Game

MF_TILESET_ALT_PAL_ROWS = {
    0x46F134: 0xD,  # 08
    0x4CCCC0: 0xD,  # 09, 40
    0x407E9C: 0xB,  # 0B, 1E
    0x498808: 0xA,  # 0E
    0x46E6F4: 0xD,  # 12
    0x46DCB4: 0xD,  # 13
    0x406B5C: 0xD,  # 19
    0x46E2F4: 0xD,  # 1B
    0x4CC5C0: 0xD,  # 1F
    0x4CC780: 0xD,  # 20
    0x4F21F8: 0xD,  # 21
    0x4F23B8: 0xD,  # 22
    0x4F27B8: 0xC,  # 28
    0x498AC8: 0xB,  # 29
    0x4F2A58: 0xC,  # 2A
    0x5352F0: 0xC,  # 2B, 2F
    0x510868: 0xD,  # 30
    0x510A28: 0xD,  # 31
    0x5355D0: 0xD,  # 34
    0x46E4B4: 0xC,  # 38
    0x46E8B4: 0xD,  # 3D
    0x54D53C: 0xB,  # 3E
    0x40719C: 0xC,  # 43
    0x46E134: 0xC,  # 46
    0x4060DC: 0xD,  # 48
    0x46EF74: 0xD,  # 54
    0x535950: 0xC,  # 56
    0x407CDC: 0xD,  # 57
    0x498648: 0xD,  # 58
    0x535B10: 0xD,  # 5B
    0x4078DC: 0xD,  # 5E
}


# Animated palettes not included here are used for multiple tileset palettes
TILESET_ANIM_PALS = {
    0x40769C: 0x01,  # Navigation room
    0x407A9C: 0x02,  # Save room
    0x406F5C: 0x03,  # Recharge room
    0x405E9C: 0x04,  # Data room
    # 0x05 Docking bay
    0x46EA74: 0x07,  # Operations deck
    # 0x08 Main elevator top/bottom
    0x46BCD4: 0x09,  # Sector 1 elevator
    0x46BF94: 0x0A,  # Sector 2 elevator
    0x46C254: 0x0B,  # Sector 3 elevator
    0x46C514: 0x0C,  # Sector 4 elevator
    0x46C7D4: 0x0D,  # Sector 5 elevator
    0x46CA94: 0x0E,  # Sector 6 elevator
    0x4F27B8: 0x0F,  # Sector 3 glowing BG3
    0x4F2A58: 0x10,  # Meltdown alarm
    0x498AC8: 0x11,  # Stabilizer rooms
    0x54D0DC: 0x12,  # Sector 6 glowing BG3
    0x406D1C: 0x13,  # Security room
    # 0x14 Electrified water
    0x46DE74: 0x15,  # Near quarantine bay
    0x4F2578: 0x16,  # Boiler control room
    0x4F1D38: 0x17,  # Main boiler
    0x46ECB4: 0x18,  # Operations deck (SA-X fight)
    0x4F1F98: 0x19,  # Main boiler
    0x46E4B4: 0x1A,  # Auxiliary power
    0x46B814: 0x1B,  # Reactor silo
    # 0x1C Restricted lab alarm
    0x46AF14: 0x1D,  # Operations room
    0x40719C: 0x1E,  # Sector 2 flashing panels 1
    0x40741C: 0x1F,  # Sector 2 flashing panels 2
    0x46D554: 0x20,  # Restricted lab
}


EXCLUDED_ENEMIES = {
    Game.MF: {
        0x11,  # SA-X (elevator)
        0x1D,  # Area banner
        0x21,  # Message box
        0x8C,  # SA-X (TRO 1)
        0x8D,  # SA-X (NOC)
        0xA0,  # SA-X (ARC)
        0xBC,  # SA-X (lab)
        0xC0,  # SA-X (boss)
        0xC1,  # SA-X monster
        0xC9,  # SA-X (Omega cutscene)
        0xCD,  # SA-X (TRO 2)
    },
    Game.ZM: {
        0x10,  # Unused
        0x11,  # Item banner
        0x8A,  # Crocomire
    },
}

ENEMY_GROUPS = {
    Game.MF: {
        "Zebesian": [0x14, 0x73, 0xA2, 0xB3, 0xB4],
        "Zeela": [0x31, 0xA3],
        "Sciser": [0x3C, 0x8B],
        "BeamCoreX": [0x56, 0x57, 0x58, 0x59],
        "Zoro": [0x5A, 0x63, 0x89, 0x8A],
        "FakeTank": [0xAD, 0xAE],
    },
    Game.ZM: {
        "Ripper": [0x16, 0x17, 0x71],
        "Zeb": [0x18, 0x19],
        "Skree": [0x1F, 0x20],
        "ChozoStatue": [
            0x22,
            0x23,
            0x24,
            0x25,
            0x26,
            0x27,
            0x28,
            0x29,
            0x2A,
            0x2B,
            0x2C,
            0x2D,
            0x2E,
            0x2F,
            0x30,
            0x31,
            0x58,
            0x59,
            0x94,
        ],
        "Viola": [0x34, 0x68, 0x69],
        "GerutaTangleVine": [0x36, 0x37, 0xB7, 0xB8, 0xB9, 0xBA],
        "Reo": [0x3F, 0x40],
        "Imago": [0x4D, 0x4F, 0x70, 0x7F, 0x86, 0x8B],
        "SpacePirate": [0x51, 0x52, 0x53, 0x54, 0x55, 0xB6, 0xC6, 0xC7],
        "Gamet": [0x56, 0x57, 0x9D, 0x9E],
        "Zebbo": [0x5B, 0x5C, 0xA1, 0xA2],
        "Rinka": [0x66, 0x7C, 0xA5, 0xC9, 0xCA, 0xCB, 0xCC, 0xCD],
        "Geega": [0x7A, 0x7B, 0x9F, 0xA0],
        "ElevatorStatue": [0x95, 0x96],
        "UnknownItemBlocks": [0xAA, 0xAB, 0xAC],
    },
}

NETTORI_EXTRA_PALS = [
    (0x36A480, 1),  # Medium health palette
    (0x36A4A0, 1),  # Low health palette
    (0x36A4C0, 3),  # Animated palette
]



import mars_patcher.constants.game_data as gd
from mars_patcher.constants.reserved_space import ReservedConstants
from mars_patcher.data import get_data_path
from mars_patcher.patching import BpsDecoder, IpsDecoder
from mars_patcher.rom import Rom


def _get_patch_path(rom: Rom, subfolder: str, filename: str) -> str:
    dir = f"{rom.game.name}_{rom.region.name}".lower()
    return get_data_path("patches", dir, subfolder, filename)


def _internal_apply_ips_patch(rom: Rom, patch_name: str, subfolder: str) -> None:
    path = _get_patch_path(rom, subfolder, patch_name)
    with open(path, "rb") as f:
        patch = f.read()
    IpsDecoder().apply_patch(patch, rom.data)


def apply_patch_in_data_path(rom: Rom, patch_name: str) -> None:
    _internal_apply_ips_patch(rom, patch_name, "")


def apply_patch_in_asm_path(rom: Rom, patch_name: str) -> None:
    _internal_apply_ips_patch(rom, patch_name, "asm")


def apply_base_patch(rom: Rom) -> None:
    path = _get_patch_path(rom, "asm", "m4rs.bps")
    with open(path, "rb") as f:
        patch = f.read()
    rom.data = BpsDecoder().apply_patch(patch, rom.data)


def disable_demos(rom: Rom) -> None:
    # b 0x8087460
    rom.write_16(0x87436, 0xE013)


def skip_door_transitions(rom: Rom) -> None:
    rom.write_32(0x69500, 0x3000BDE)
    rom.write_8(0x694E2, 0xC)


def stereo_default(rom: Rom) -> None:
    apply_patch_in_data_path(rom, "stereo_default.ips")


def disable_sounds(rom: Rom, start: int, end: int, exclude: set[int] = set()) -> None:
    sound_data_addr = gd.sound_data_entries(rom)
    for idx in range(start, end):
        if idx not in exclude:
            addr = sound_data_addr + idx * 8
            rom.write_8(rom.read_ptr(addr), 0)


def disable_music(rom: Rom) -> None:
    # Exclude jingles
    exclude = {
        16,  # Major obtained
        17,  # Loading save
        20,  # Minor obtained
        59,  # Event
    }
    disable_sounds(rom, 0, 100, exclude)


def disable_sound_effects(rom: Rom) -> None:
    disable_sounds(rom, 100, gd.sound_count(rom))


def change_missile_limit(rom: Rom, limit: int) -> None:
    rom.write_8(rom.read_ptr(ReservedConstants.MISSILE_LIMIT_ADDR), limit)


def apply_unexplored_map(rom: Rom) -> None:
    apply_patch_in_asm_path(rom, "unhidden_map.ips")


def apply_pbs_without_bombs(rom: Rom) -> None:
    apply_patch_in_asm_path(rom, "bombless_pbs.ips")


def apply_reveal_hidden_tiles(rom: Rom) -> None:
    rom.write_8(rom.read_ptr(ReservedConstants.REVEAL_HIDDEN_TILES_ADDR), 1)


def apply_reveal_unexplored_doors(rom: Rom) -> None:
    apply_patch_in_asm_path(rom, "unhidden_map_doors.ips")


def apply_accessibility_patch(rom: Rom) -> None:
    apply_patch_in_asm_path(rom, "accessibility.ips")



import json
from enum import Enum
from functools import cache

from mars_patcher.constants.game_data import character_widths, file_screen_text_ptrs
from mars_patcher.data import get_data_path
from mars_patcher.rom import Region, Rom

SPACE_CHAR = 0x40
SPACE_TAG = 0x8000
NEXT = 0xFD00
NEWLINE = 0xFE00
END = 0xFF00
VALUE_MARKUP_TAG = {
    "SPACE": (SPACE_TAG, 8),
    "COLOR": (0x8100, 8),
    "SPEED": (0x8200, 8),
    "INDENT": (0x8300, 8),
    "PLAY_SOUND": (0x9000, 12),
    "STOP_SOUND": (0xA000, 12),
    "WAIT": (0xE100, 8),
}
BREAKING_CHARS = {SPACE_CHAR, NEXT, NEWLINE}
NEWLINE_CHARS = {NEXT, NEWLINE}

KANJI_START = 0x4A0
KANJI_WIDTH = 10
MAX_LINE_WIDTH = 224


class Language(Enum):
    JAPANESE_KANJI = 0
    JAPANESE_HIRAGANA = 1
    ENGLISH = 2
    GERMAN = 3
    FRENCH = 4
    ITALIAN = 5
    SPANISH = 6


class MessageType(Enum):
    """Message types for encoding."""

    ONE_LINE = 0
    """Used for text that can only span one line (ex: location banner)."""
    TWO_LINE = 1
    """Used for text that can only span two lines (ex: item/event message,
    pause screen objective)."""
    CONTINUOUS = 2
    """Used for text where the A button can advance text."""


@cache
def get_char_map(region: Region) -> dict[str, int]:
    path = get_data_path("char_map_mf.json")
    with open(path, encoding="utf-8") as f:
        sections = json.load(f)
    char_map: dict[str, int] = {}
    for section in sections:
        if region.name in section["regions"]:
            char_map.update(section["chars"])
    char_map["\n"] = NEWLINE
    return char_map


def parse_value_markup_tag(tag: str) -> int | None:
    """Used to try parsing a markup tag with an assignable value.
    Returns the resulting character value, or None if not a markup tag."""
    items = tag.split("=", 1)
    if len(items) != 2:
        return None
    label, value_str = items
    if label in VALUE_MARKUP_TAG:
        char_val, bits = VALUE_MARKUP_TAG[label]
        value = int(value_str, 16)
        if value < 0 or value >= (1 << bits):
            raise ValueError(f"Value {value} is not valid for {label}")
        return char_val | value
    raise ValueError(f"Invalid value markup tag '{tag}'")


def get_char_width(rom: Rom, char_widths_addr: int, char_val: int) -> int:
    if char_val >= 0x8000:
        return 0
    if char_val < KANJI_START:
        return rom.read_8(char_widths_addr + char_val)
    return KANJI_WIDTH


def center_text(rom: Rom, char_vals: list[int], max_width: int) -> None:
    char_widths_addr = character_widths(rom)
    line_start = 0
    line_width = 0
    index = 0
    while index < len(char_vals):
        char_val = char_vals[index]
        index += 1
        line_width += get_char_width(rom, char_widths_addr, char_val)
        if char_val in NEWLINE_CHARS or index == len(char_vals):
            if line_width > 0:
                assert line_width <= max_width
                space_val = SPACE_TAG + (max_width - line_width) // 2
                char_vals.insert(line_start, space_val)
                index += 1
                line_width = 0
            line_start = index


def encode_text(
    rom: Rom,
    message_type: MessageType,
    string: str,
    max_width: int = MAX_LINE_WIDTH,
    centered: bool = False,
) -> list[int]:
    char_map = get_char_map(rom.region)
    char_widths_addr = character_widths(rom)
    text: list[int] = []
    line_width = 0
    line_number = 0

    prev_break: int | None = None
    width_since_break = 0
    escaped = False
    markup_tag: list[str] | None = None

    for char in string:
        if not escaped:
            # Check for escaped character
            if char == "\\":
                if markup_tag is not None:
                    raise ValueError(f'Escaped character in markup tag:\n"{string}"')
                escaped = True
                continue
            if markup_tag is None:
                # Check for start of markup tag
                if char == "[":
                    markup_tag = []
                    continue
            else:
                # Check for end of markup tag
                if char == "]":
                    tag_str = "".join(markup_tag)
                    # Check if markup tag with assignable value
                    char_val = parse_value_markup_tag(tag_str)
                    if char_val is None:
                        # Check if normal markup tag
                        char_val = char_map.get(f"[{tag_str}]")
                        if char_val is None:
                            raise ValueError(f"Invalid markup tag '{tag_str}'")
                    text.append(char_val)
                    markup_tag = None
                else:
                    markup_tag.append(char)
                continue
        else:
            escaped = False

        char_val = char_map[char]
        char_width = get_char_width(rom, char_widths_addr, char_val)
        line_width += char_width
        width_since_break += char_width

        if char_val in BREAKING_CHARS:
            prev_break = len(text)
            width_since_break = 0
            if char_val in NEWLINE_CHARS:
                line_width = 0
                line_number += 1

        extra_char = None

        if line_width > max_width:
            if message_type == MessageType.ONE_LINE:
                raise ValueError(f'String does not fit on one line:\n"{string}"')
            if width_since_break > max_width:
                raise ValueError(f'Word does not fit on one line:\n"{string}"')
            line_width = width_since_break
            line_number += 1
            extra_char = NEWLINE

        if line_number > 1:
            match message_type:
                case MessageType.CONTINUOUS:
                    line_number = 0
                    extra_char = NEXT
                case MessageType.TWO_LINE:
                    # Limited to 2 lines, trim any other characters
                    break

        if extra_char is not None:
            if prev_break is not None:
                if len(text) <= prev_break:
                    text.append(extra_char)
                    continue
                else:
                    text[prev_break] = extra_char
                prev_break = None
            else:
                text.append(extra_char)

        text.append(char_val)

    if markup_tag is not None:
        raise ValueError(f'Unclosed markup tag:\n"{string}"')

    if message_type == MessageType.ONE_LINE and (NEXT in text or NEWLINE in text):
        raise ValueError(f'String cannot have newlines:\n"{string}"')

    if centered:
        center_text(rom, text, max_width)

    if message_type == MessageType.TWO_LINE and NEWLINE not in text:
        # Two line messages MUST have two lines, append NEWLINE if none exists
        text.append(NEWLINE)

    text.append(END)
    return text


def write_seed_hash(rom: Rom, seed_hash: str) -> None:
    char_map = get_char_map(rom.region)
    lang_ptrs = file_screen_text_ptrs(rom)
    for lang in Language:
        # Get address of first text entry
        text_ptrs = rom.read_ptr(lang_ptrs + lang.value * 4)
        addr = rom.read_ptr(text_ptrs)
        # Find newline after "SAMUS DATA"
        try:
            line_len = next(i for i in range(20) if rom.read_16(addr + i * 2) == NEWLINE)
        except StopIteration:
            raise ValueError("Invalid file screen text data")
        pad_left = (line_len - 8) // 2
        pad_right = line_len - 8 - pad_left
        # Overwrite with seed hash
        string = (" " * pad_left) + seed_hash + (" " * pad_right)
        for i, c in enumerate(string):
            rom.write_16(addr + i * 2, char_map[c])



from typing import Annotated, TypeAlias

from mars_patcher.auto_generated_types import Areaid, Typeu8

AreaId: TypeAlias = Areaid
RoomId: TypeAlias = Typeu8

AreaRoomPair = tuple[AreaId, RoomId]

MinimapId: TypeAlias = Annotated[int, "0 <= value < 10"]



import random

from mars_patcher.constants.enemies import ENEMY_TYPES, EnemyType
from mars_patcher.constants.game_data import sprite_vram_sizes, spriteset_count, spriteset_ptrs
from mars_patcher.rom import Rom


def randomize_enemies(rom: Rom) -> None:
    # Setup enemy types dictionary
    enemy_types = {k: v[1] for k, v in ENEMY_TYPES.items()}

    # Get graphics info for each enemy
    size_addr = sprite_vram_sizes(rom)
    gfx_rows = {}
    for en_id in enemy_types:
        size = rom.read_32(size_addr + (en_id - 0x10) * 4)
        gfx_rows[en_id] = size // 0x800

    # Get replacement pools
    replacements: dict[EnemyType, list[int]] = {t: [] for t in EnemyType}
    for en_id, en_type in enemy_types.items():
        replacements[en_type].append(en_id)
        if en_type == EnemyType.CRAWLING:
            replacements[EnemyType.GROUND].append(en_id)
            replacements[EnemyType.CEILING].append(en_id)
            replacements[EnemyType.GROUND_CEILING].append(en_id)
            replacements[EnemyType.WALL].append(en_id)
        elif en_type == EnemyType.GROUND_CEILING:
            replacements[EnemyType.GROUND].append(en_id)
            replacements[EnemyType.CEILING].append(en_id)
        # Ground, Ceiling, Wall, and Flying cannot replace others

    # Randomize spritesets
    ss_ptrs = spriteset_ptrs(rom)
    for i in range(spriteset_count(rom)):
        spriteset_addr = rom.read_ptr(ss_ptrs + i * 4)
        used_gfx_rows: dict[int, int] = {}
        spriteset = get_spriteset(rom, spriteset_addr)
        for j, (en_id, gfx_row) in enumerate(spriteset):
            # Skip enemies that aren't randomized
            if en_id not in enemy_types:
                continue

            # Check if sprite shares graphics with one that's never randomized
            if any(
                en_id != other_id and gfx_row == other_row and other_id not in enemy_types
                for other_id, other_row in spriteset
            ):
                continue

            # Check if sprite shares graphics with one that's already randomized
            addr = spriteset_addr + (j * 2)
            if gfx_row in used_gfx_rows:
                new_id = used_gfx_rows[gfx_row]
                rom.write_8(addr, new_id)
                continue

            # Choose randomly and assign
            en_type = enemy_types[en_id]
            row_count = gfx_rows[en_id]
            candidates = replacements[en_type]
            random.shuffle(candidates)
            for new_id in candidates:
                new_row_count = gfx_rows[new_id]
                # New enemy must use same or fewer graphics rows
                if new_row_count <= row_count:
                    rom.write_8(addr, new_id)
                    used_gfx_rows[gfx_row] = new_id
                    break


def get_spriteset(rom: Rom, addr: int) -> list[tuple[int, int]]:
    """Returns a list of (sprite ID, graphics row) tuples in the spriteset
    at the provided address."""
    spriteset: list[tuple[int, int]] = []
    for _ in range(0xF):
        en_id = rom.read_8(addr)
        if en_id == 0:
            break
        gfx_row = rom.read_8(addr + 1)
        spriteset.append((en_id, gfx_row))
        addr += 2
    return spriteset



from mars_patcher.constants.game_data import tileset_entries
from mars_patcher.rom import Rom


class Tileset:
    def __init__(self, rom: Rom, id: int):
        self.rom = rom
        self.addr = tileset_entries(rom) + id * 0x14

    def rle_tilemap_addr(self) -> int:
        return self.rom.read_ptr(self.addr + 0xC)



from collections.abc import Sequence

import mars_patcher.constants.game_data as gd
from mars_patcher.auto_generated_types import (
    MarsschemaElevatorconnections,
    MarsschemaSectorshortcuts,
    Validelevatorbottoms,
    Validelevatortops,
)
from mars_patcher.constants.main_hub_numbers import (
    MAIN_HUB_CENTER_ROOM,
    MAIN_HUB_CENTER_SMALL_NUM_COORDS_1,
    MAIN_HUB_CENTER_SMALL_NUM_COORDS_2,
    MAIN_HUB_ELE_DOORS,
    MAIN_HUB_ELE_ROOM_LARGE_NUM_COORD,
    MAIN_HUB_ELE_ROOM_SMALL_NUM_COORDS,
    MAIN_HUB_ELE_ROOMS,
    MAIN_HUB_GFX_ADDR,
    MAIN_HUB_LARGE_NUM_BLOCKS,
    MAIN_HUB_SMALL_NUM_BLOCK,
    MAIN_HUB_TILEMAP_ADDR,
)
from mars_patcher.data import get_data_path
from mars_patcher.minimap import Minimap
from mars_patcher.rom import Game, Rom
from mars_patcher.room_entry import BlockLayer, RoomEntry

# Area ID, Room ID, Is area connection
ELEVATOR_TOPS = {
    "OperationsDeckTop": (0, 0x1A, False),
    "MainHubToSector1": (0, 0x43, True),
    "MainHubToSector2": (0, 0x44, True),
    "MainHubToSector3": (0, 0x45, True),
    "MainHubToSector4": (0, 0x46, True),
    "MainHubToSector5": (0, 0x47, True),
    "MainHubToSector6": (0, 0x48, True),
    "MainHubTop": (0, 0x5E, False),
    "HabitationDeckTop": (0, 0xB2, False),
    "Sector1ToRestrictedLab": (1, 0x41, True),
}

ELEVATOR_BOTTOMS = {
    "OperationsDeckBottom": (0, 0x19, False),
    "MainHubBottom": (0, 0x32, False),
    "RestrictedLabToSector1": (0, 0x9A, True),
    "HabitationDeckBottom": (0, 0xB1, False),
    "Sector1ToMainHub": (1, 0x00, True),
    "Sector2ToMainHub": (2, 0x00, True),
    "Sector3ToMainHub": (3, 0x00, True),
    "Sector4ToMainHub": (4, 0x00, True),
    "Sector5ToMainHub": (5, 0x00, True),
    "Sector6ToMainHub": (6, 0x00, True),
}

# (Area ID, Dest area): Door ID
SHORTCUT_LEFT_DOORS = [0x6B, 0x7F, 0x59, 0x6C, 0x02, 0x51]
SHORTCUT_RIGHT_DOORS = [0x68, 0x82, 0x56, 0x6A, 0x53, 0x54]

SHORTCUT_LEFT_NUM_COORD = (6, 3)
SHORTCUT_RIGHT_NUM_COORD = (9, 3)
SHORTCUT_NUM_X_OFFSET = 3
SHORTCUT_NUM_BLOCKS = [0x101, 0x104, 0xE5, 0xE6, 0xE7, 0xE8]

DOOR_TYPE_AREA_CONN = 1
DOOR_TYPE_NO_HATCH = 2


class Connections:
    """Class for handling elevator shuffle and sector shortcut shuffle."""

    def __init__(self, rom: Rom):
        self.rom = rom
        self.area_doors_ptrs = gd.area_doors_ptrs(rom)
        self.area_conns_addr = gd.area_connections(rom)
        self.area_conns_count = gd.area_connections_count(rom)

    def set_elevator_connections(self, data: MarsschemaElevatorconnections) -> None:
        # Repoint area connections data
        size = self.area_conns_count * 3
        # Reserve space for 8 more area connections
        new_size = size + 8 * 3
        ac_addr = self.rom.reserve_free_space(new_size)
        self.rom.copy_bytes(self.area_conns_addr, ac_addr, size)
        # TODO: Move constant
        self.rom.write_ptr(0x6945C, ac_addr)
        self.area_conns_addr = ac_addr

        # Connect tops to bottoms
        pairs_top = data["ElevatorTops"]
        self.connect_elevators(ELEVATOR_TOPS, ELEVATOR_BOTTOMS, pairs_top)
        # Connect bottoms to tops
        pairs_bottom = data["ElevatorBottoms"]
        self.connect_elevators(ELEVATOR_BOTTOMS, ELEVATOR_TOPS, pairs_bottom)
        if self.rom.game == Game.MF:
            # Update area number tiles in main hub rooms
            self.fix_main_hub_tiles()
            # Remove area numbers from Main Deck minimap
            self.remove_main_deck_minimap_area_nums()

    def set_shortcut_connections(self, data: MarsschemaSectorshortcuts) -> None:
        for i, dst_area in enumerate(data["LeftAreas"]):
            self.connect_shortcuts(i + 1, dst_area, True)
        for i, dst_area in enumerate(data["RightAreas"]):
            self.connect_shortcuts(i + 1, dst_area, False)

    def connect_shortcuts(self, area: int, dst_area: int, left: bool) -> None:
        # Connect doors and update area connection
        if left:
            src_list = SHORTCUT_LEFT_DOORS
            dst_list = SHORTCUT_RIGHT_DOORS
            x, y = SHORTCUT_LEFT_NUM_COORD
            left_area, right_area = dst_area, area
        else:
            src_list = SHORTCUT_RIGHT_DOORS
            dst_list = SHORTCUT_LEFT_DOORS
            x, y = SHORTCUT_RIGHT_NUM_COORD
            left_area, right_area = area, dst_area
        door = src_list[area - 1]
        dst_door = dst_list[dst_area - 1]
        self.connect_doors(area, door, dst_area, dst_door)
        self.connect_areas(area, door, dst_area, True)

        # Update area numbers on BG1
        addr = self.rom.read_ptr(self.area_doors_ptrs + area * 4) + door * 0xC
        room = self.rom.read_8(addr + 1)
        room_entry = RoomEntry(self.rom, area, room)
        with room_entry.load_bg1() as bg1:
            block = SHORTCUT_NUM_BLOCKS[left_area - 1]
            bg1.set_block_value(x, y, block)
            bg1.set_block_value(x, y + 1, block + 0x10)
            block = SHORTCUT_NUM_BLOCKS[right_area - 1]
            x += SHORTCUT_NUM_X_OFFSET
            bg1.set_block_value(x, y, block)
            bg1.set_block_value(x, y + 1, block + 0x10)

    def connect_elevators(
        self,
        src_dict: dict,
        dst_dict: dict,
        pairs: dict[Validelevatortops, Validelevatorbottoms]
        | dict[Validelevatorbottoms, Validelevatortops],
    ) -> None:
        for src_name, dst_name in pairs.items():
            src_area, src_door, in_list = src_dict[src_name]
            dst_area, dst_door, _ = dst_dict[dst_name]
            # Modify door entry
            self.connect_doors(src_area, src_door, dst_area, dst_door)
            # Modify area connection
            self.connect_areas(src_area, src_door, dst_area, in_list)

    def connect_doors(self, src_area: int, src_door: int, dst_area: int, dst_door: int) -> None:
        addr = self.rom.read_ptr(self.area_doors_ptrs + src_area * 4) + src_door * 0xC
        # Fix door type
        props = self.rom.read_8(addr)
        door_type = DOOR_TYPE_AREA_CONN if src_area != dst_area else DOOR_TYPE_NO_HATCH
        self.rom.write_8(addr, props & 0xF0 | door_type)
        # Set destination door
        self.rom.write_8(addr + 6, dst_door)

    def connect_areas(self, src_area: int, src_door: int, dst_area: int, in_list: bool) -> None:
        rom = self.rom
        same_area = src_area == dst_area
        if in_list:
            # Find existing area connection
            for i in range(self.area_conns_count):
                addr = self.area_conns_addr + i * 3
                if rom.read_8(addr) == src_area and rom.read_8(addr + 1) == src_door:
                    if same_area:
                        # Make entry blank
                        rom.write_8(addr, 0)
                        rom.write_8(addr + 1, 0)
                        rom.write_8(addr + 2, 0)
                    else:
                        rom.write_8(addr + 2, dst_area)
                    return
            raise ValueError(f"Area connection not found for Area {src_area} Door {src_door:02X}")
        elif not same_area:
            addr = self.area_conns_addr + self.area_conns_count * 3
            rom.write_8(addr, src_area)
            rom.write_8(addr + 1, src_door)
            rom.write_8(addr + 2, dst_area)
            self.area_conns_count += 1

    def fix_main_hub_tiles(self) -> None:
        # Get areas that the 6 elevators go to
        ele_areas = [0 for _ in MAIN_HUB_ELE_DOORS]
        for i in range(self.area_conns_count):
            addr = self.area_conns_addr + i * 3
            # Skip if not main deck
            if self.rom.read_8(addr) != 0:
                continue
            door = self.rom.read_8(addr + 1)
            for j, ele_door in enumerate(MAIN_HUB_ELE_DOORS):
                if door == ele_door:
                    ele_areas[j] = self.rom.read_8(addr + 2)
                    break

        # Write new graphics and tilemap
        path = get_data_path("main_hub.gfx.lz")
        with open(path, "rb") as f:
            gfx = f.read()
        self.rom.write_bytes(MAIN_HUB_GFX_ADDR, gfx)
        path = get_data_path("main_hub_tilemap.bin")
        with open(path, "rb") as f:
            tilemap = f.read()
        self.rom.write_bytes(MAIN_HUB_TILEMAP_ADDR + 2, tilemap)

        # Overwrite numbers on BG2
        # Central room
        room_entry = RoomEntry(self.rom, 0, MAIN_HUB_CENTER_ROOM)
        with room_entry.load_bg2() as bg2:
            self._write_main_hub_small_nums(bg2, MAIN_HUB_CENTER_SMALL_NUM_COORDS_1, ele_areas)
            self._write_main_hub_small_nums(bg2, MAIN_HUB_CENTER_SMALL_NUM_COORDS_2, ele_areas)
        # Elevator rooms
        large_x, large_y = MAIN_HUB_ELE_ROOM_LARGE_NUM_COORD
        for i, room in enumerate(MAIN_HUB_ELE_ROOMS):
            coords = MAIN_HUB_ELE_ROOM_SMALL_NUM_COORDS[i]
            room_entry = RoomEntry(self.rom, 0, room)
            with room_entry.load_bg2() as bg2:
                self._write_main_hub_small_nums(bg2, coords, ele_areas)
                block = MAIN_HUB_LARGE_NUM_BLOCKS[ele_areas[i]]
                bg2.set_block_value(large_x, large_y, block)
                bg2.set_block_value(large_x, large_y + 1, block + 0x10)

    def _write_main_hub_small_nums(
        self, bg2: BlockLayer, coords: Sequence[tuple[int, int] | None], ele_areas: list[int]
    ) -> None:
        for area, coord in enumerate(coords):
            if coord is None:
                continue
            block = MAIN_HUB_SMALL_NUM_BLOCK + ele_areas[area] * 0x10
            if area % 2 == 0:
                block += 2
            x, y = coord
            bg2.set_block_value(x, y, block)
            bg2.set_block_value(x + 1, y, block + 1)

    def remove_main_deck_minimap_area_nums(self) -> None:
        with Minimap(self.rom, 0) as minimap:
            minimap.set_tile_value(0x2, 0x11, 0xA0, 0)  # 5
            minimap.set_tile_value(0x3, 0x10, 0xA0, 0)  # 3
            minimap.set_tile_value(0x4, 0x0F, 0xA0, 0)  # 1
            minimap.set_tile_value(0x8, 0x0F, 0xA0, 0)  # 2
            minimap.set_tile_value(0x9, 0x10, 0xA0, 0)  # 4
            minimap.set_tile_value(0xA, 0x11, 0xA0, 0)  # 6



from enum import Enum


class EnemyType(Enum):
    CRAWLING = 1
    GROUND = 2
    CEILING = 3
    GROUND_CEILING = 4
    WALL = 5
    FLYING = 6


ENEMY_TYPES = {
    0x12: ("Hornoad", EnemyType.GROUND),
    0x13: ("Halzyn", EnemyType.FLYING),
    0x14: ("Zebesian (wall)", EnemyType.WALL),
    0x15: ("Hornoad spawner", EnemyType.GROUND),
    0x18: ("Moto", EnemyType.GROUND),
    0x1B: ("Yameba", EnemyType.FLYING),
    0x31: ("Zeela", EnemyType.CRAWLING),
    0x33: ("Skree (unused)", EnemyType.CEILING),
    0x37: ("Zombie", EnemyType.GROUND),
    0x39: ("Geemer", EnemyType.CRAWLING),
    0x3B: ("Waver", EnemyType.FLYING),
    0x3C: ("Sciser", EnemyType.CRAWLING),
    0x3D: ("Sidehopper", EnemyType.GROUND_CEILING),
    0x3E: ("Dessgeega", EnemyType.GROUND_CEILING),
    0x5A: ("Zoro", EnemyType.CRAWLING),
    0x5B: ("Kihunter (flying)", EnemyType.FLYING),
    0x5C: ("Kihunter (ground)", EnemyType.GROUND),
    0x5E: ("Reo", EnemyType.FLYING),
    0x60: ("Namihe", EnemyType.WALL),
    0x61: ("Fune", EnemyType.WALL),
    0x63: ("Blue zoro", EnemyType.CRAWLING),
    0x64: ("Geruda", EnemyType.FLYING),
    0x6A: ("Skultera (large)", EnemyType.FLYING),
    0x6B: ("Skultera (small, cannot become large)", EnemyType.FLYING),
    0x6C: ("Sova", EnemyType.CRAWLING),
    0x6D: ("Yard", EnemyType.CRAWLING),
    0x6E: ("Evir", EnemyType.FLYING),
    0x6F: ("Bull", EnemyType.FLYING),
    0x70: ("Memu", EnemyType.FLYING),
    0x71: ("Geruboss", EnemyType.CEILING),
    0x72: ("Choot", EnemyType.GROUND),
    0x73: ("Zebesian (ground)", EnemyType.GROUND),
    0x88: ("Ripper", EnemyType.FLYING),
    0xA3: ("Red Zeela", EnemyType.CRAWLING),
    0xA4: ("Owtch", EnemyType.CRAWLING),
    0xA8: ("Genesis", EnemyType.GROUND),
    0xA9: ("Puyo", EnemyType.GROUND),
    0xAD: ("Fake energy tank", EnemyType.GROUND),
    0xAE: ("Fake missile tank", EnemyType.GROUND),
    0xB3: ("Zebesian (aqua)", EnemyType.FLYING),
    0xB4: ("Zebesian (pre-aqua)", EnemyType.GROUND),
    0xB5: ("Skultera (small)", EnemyType.FLYING),
    0xBD: ("Powamp", EnemyType.FLYING),
    0xBE: ("Zozoro", EnemyType.CRAWLING),
}



MIN_MATCH_SIZE = 3
MIN_WINDOW_SIZE = 1
MAX_MATCH_SIZE = (1 << 4) - 1 + MIN_MATCH_SIZE
MAX_WINDOW_SIZE = (1 << 12) - 1 + MIN_WINDOW_SIZE


def decomp_rle(input: bytes, idx: int) -> tuple[bytearray, int]:
    """
    Decompresses RLE data and returns it with the size of the compressed data.
    """
    src_start = idx
    passes = bytearray()
    half: int | None = None
    # For each pass
    for p in range(2):
        if p == 1:
            half = len(passes)
        num_bytes = input[idx]
        idx += 1
        while True:
            amount = None
            compare = None
            if num_bytes == 1:
                amount = input[idx]
                compare = 0x80
            else:
                # num_bytes == 2
                amount = (input[idx] << 8) | input[idx + 1]
                compare = 0x8000
            idx += num_bytes

            if amount == 0:
                break

            if (amount & compare) != 0:
                # Compressed
                amount %= compare
                val = input[idx]
                idx += 1
                while amount > 0:
                    passes.append(val)
                    amount -= 1
            else:
                # Uncompressed
                while amount > 0:
                    passes.append(input[idx])
                    idx += 1
                    amount -= 1

    # Each pass must be equal length
    if half is None:
        raise ValueError("half was not assigned")
    if len(passes) != half * 2:
        raise ValueError("Passes are not equal length")

    # Combine passes to get output
    output = bytearray(len(passes))
    for i in range(half):
        output[i * 2] = passes[i]
        output[i * 2 + 1] = passes[half + i]

    # Return bytes and compressed size
    comp_size = idx - src_start
    return output, comp_size


def comp_rle(input: bytes) -> bytearray:
    """
    Compresses data using RLE.
    """

    # Inner helper functions
    def write_len(arr: bytearray, val: int, size: int) -> None:
        if size == 0:
            arr.append(val)
        else:
            arr.append(val >> 8)
            arr.append(val & 0xFF)

    def add_unique(arr: bytearray, unique: bytearray, size: int) -> None:
        write_len(arr, len(unique), size)
        arr += unique
        unique.clear()

    output = bytearray()
    # Do two passes for low and high bytes
    for p in range(2):
        # Get counts of consecutive values
        values = bytearray()
        counts = []

        prev = input[p]
        values.append(prev)
        count = 1

        for i in range(p + 2, len(input), 2):
            val = input[i]
            if val == prev:
                count += 1
            else:
                values.append(val)
                counts.append(count)
                prev = val
                count = 1
        counts.append(count)

        # Try each read length (1 or 2)
        shortest: bytearray | None = None
        for r in range(2):
            temp = bytearray()
            min_run_len = 3 + r
            flag = 0x80 << (8 * r)
            max_run_len = flag - 1
            unique = bytearray()

            # Write number of bytes to read
            temp.append(r + 1)

            # For each value and its count
            for i in range(len(values)):
                count = counts[i]
                # If the value's count is long enough for a run
                if count >= min_run_len:
                    # If the value is preceded by unique values
                    if len(unique) > 0:
                        add_unique(temp, unique, r)
                    # Add run length and value (multiple times if over max run length)
                    while count > 0:
                        curr_len = min(count, max_run_len)
                        len_flag = curr_len + flag
                        write_len(temp, len_flag, r)
                        temp.append(values[i])
                        count -= curr_len
                # If the value's count is too short for a run
                else:
                    # If the total count would be too long for a run
                    if len(unique) + count > max_run_len:
                        add_unique(temp, unique, r)
                    for _ in range(count):
                        unique.append(values[i])
            # Check if there were unique values at the end
            if len(unique) > 0:
                add_unique(temp, unique, r)
            # Write ending zero(s)
            temp.append(0)
            if r == 1:
                temp.append(0)
            if shortest is None or len(temp) < len(shortest):
                shortest = temp
        if shortest is None:
            raise ValueError("shortest was not assigned")
        output += shortest
    return output


def decomp_lz77(input: bytes, idx: int) -> tuple[bytearray, int]:
    """Decompresses LZ77 data and returns it with the size of the compressed data."""
    # Check for 0x10 flag
    if input[idx] != 0x10:
        raise ValueError("Missing 0x10 flag")

    # Get length of decompressed data
    remain = input[idx + 1] | (input[idx + 2] << 8) | (input[idx + 3] << 16)
    output = bytearray([0] * remain)

    # Check for valid data size
    if remain == 0:
        raise ValueError("Invalid data size")

    start = idx
    idx += 4
    dst = 0

    # Decompress
    while True:
        cflag = input[idx]
        idx += 1

        for _ in range(8):
            if (cflag & 0x80) == 0:
                # Uncompressed
                output[dst] = input[idx]
                idx += 1
                dst += 1
                remain -= 1
            else:
                # Compressed
                amount_to_copy = (input[idx] >> 4) + MIN_MATCH_SIZE
                window = ((input[idx] & 0xF) << 8) + input[idx + 1] + MIN_WINDOW_SIZE
                idx += 2
                remain -= amount_to_copy

                for _ in range(amount_to_copy):
                    output[dst] = output[dst - window]
                    dst += 1

            if remain <= 0:
                if remain < 0:
                    raise ValueError("Too many bytes copied at end")
                comp_size = idx - start
                return output, comp_size
            cflag <<= 1


def comp_lz77(input: bytes) -> bytearray:
    """Compresses data using LZ77."""
    length = len(input)
    idx = 0
    longest_matches = _find_longest_matches(input)

    # Write start of data
    output = bytearray()
    output.append(0x10)
    output.append(length & 0xFF)
    output.append((length >> 8) & 0xFF)
    output.append(length >> 16)

    while idx < length:
        # Get index of new compression flag
        flag = len(output)
        output.append(0)

        for i in range(8):
            # Find longest match at current position
            _match = longest_matches.get(idx)
            if _match is not None:
                # Compressed
                match_idx, match_len = _match
                match_offset = idx - match_idx - MIN_WINDOW_SIZE
                output.append(((match_len - MIN_MATCH_SIZE) << 4) | (match_offset >> 8))
                output.append(match_offset & 0xFF)
                output[flag] |= 0x80 >> i
                idx += match_len
            else:
                # Uncompressed
                output.append(input[idx])
                idx += 1

            # Check if at end
            if idx >= length:
                return output

    raise RuntimeError("LZ77 compression error")


def _find_longest_matches(input: bytes) -> dict[int, tuple[int, int]]:
    length = len(input)
    triplets: dict[int, list[int]] = {}
    longest_matches: dict[int, tuple[int, int]] = {}

    for i in range(length - 2):
        # Get triplet at current position
        triplet = input[i] | (input[i + 1] << 8) | (input[i + 2] << 16)

        # Check if triplet has no match
        indexes = triplets.get(triplet)
        if indexes is None:
            triplets[triplet] = [i]
            continue

        window_start = max(i - MAX_WINDOW_SIZE, 0)
        max_size = min(MAX_MATCH_SIZE, length - i)
        longest_len = 0
        longest_idx = -1

        # Skip first index if one byte behind current position
        j = len(indexes) - 1
        if indexes[j] >= i - 1:
            j -= 1

        # Try each index to find the longest match
        while j >= 0:
            idx = indexes[j]
            # Stop if past window
            if idx < window_start:
                break

            # Find length of match
            match_len = MIN_MATCH_SIZE
            while match_len < max_size:
                if input[idx + match_len] != input[i + match_len]:
                    break
                match_len += 1

            # Update longest match
            if match_len > longest_len:
                longest_len = match_len
                longest_idx = idx

                # Stop looking if max size
                if longest_len == max_size:
                    break

            j -= 1

        indexes.append(i)
        if longest_len >= MIN_MATCH_SIZE:
            longest_matches[i] = (longest_idx, longest_len)

    return longest_matches



from mars_patcher.auto_generated_types import MarsschemaTitletextItem
from mars_patcher.constants.reserved_space import ReservedConstants
from mars_patcher.rom import Rom

TITLE_TEXT_POINTER_ADDR = ReservedConstants.TITLESCREEN_TEXT_POINTERS_POINTER_ADDR
MAX_LENGTH = 30
MAX_LINES = 14


class TitleScreenText:
    def __init__(self, text: str | None = None, line_num: int | None = None):
        self.line_num = line_num
        self.text = text

    @classmethod
    def from_json(cls, data: MarsschemaTitletextItem) -> "TitleScreenText":
        text = data.get("Text", None)
        line_num = data.get("LineNum", None)
        return TitleScreenText(text, line_num)


def write_title_text(rom: Rom, lines: list[MarsschemaTitletextItem]) -> None:
    if len(lines) > MAX_LINES:
        raise ValueError("Maximum number of title-screen lines exceeded.")

    for line in lines:
        write_title_text_line(rom, line)


def write_title_text_line(rom: Rom, line: MarsschemaTitletextItem) -> None:
    if len(line["Text"]) > 30:
        raise ValueError(f'String for title-screen text exceeds 30 characters.\n"{line["Text"]}"')
    text_pointers = rom.read_ptr(ReservedConstants.TITLESCREEN_TEXT_POINTERS_POINTER_ADDR)
    addr = rom.read_ptr(text_pointers + (line["LineNum"] * 4))
    rom.write_bytes(addr, line["Text"].encode("ascii"))
